// TClosure.h -- Simulation of Functional Closures in C++
// DM/MCFA  08/04
//
#ifndef __TCLOSURE_H__
#define __TCLOSURE_H__

extern "C" {
#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/fail.h>
#include <caml/custom.h>
#include <caml/intext.h>
};

class TClosure
{
public:
  TClosure() {};
  virtual ~TClosure() {};
  
  virtual void operator()() = 0;
};

#ifdef __TClosure_Instantiate__
void safe_call(TClosure &closure)
{
  try {
    closure();
  } catch(const char* msg) {
    failwith((char*)msg);
  }
}
#else
extern void safe_call(TClosure &closure);
#endif

// ------------------------------------------------
// Internal macros
//
#define __VClosureHead()				\
  struct closure: public TClosure			\
  {							\
    value *m_pparm;					\
    value  m_result;					\
    closure(value *pparm)				\
      :m_pparm(pparm)					\
      {}						\

#define __VClosureArgs1(arg)			\
  enum {arg};

#define __VClosureArgs2(arg1,arg2)		\
  enum {arg1,arg2};

#define __VClosureArgs3(arg1,arg2,arg3)		\
  enum {arg1,arg2,arg3};

#define __VClosureArgs4(arg1,arg2,arg3,arg4)	\
  enum {arg1,arg2,arg3,arg4};

#define __VClosureArgs5(arg1,arg2,arg3,arg4,arg5)	\
  enum {arg1,arg2,arg3,arg4,arg5};

#define __VClosureTail(name,body,env)			\
  void operator()()					\
  { body }						\
  } name(env);

// ------------------------------------------------------
// API Level Macros

#define VClosure(name,body,env)	     \
  __VClosureHead()		     \
    __VClosureTail(name,body,env)

#define VClosure1(name,arg,body,env) \
  __VClosureHead()		     \
    __VClosureArgs1(arg)		     \
       __VClosureTail(name,body,env)

#define VClosure2(name,arg1,arg2,body,env)	\
  __VClosureHead()				\
    __VClosureArgs2(arg1,arg2)			\
       __VClosureTail(name,body,env)

#define VClosure3(name,arg1,arg2,arg3,body,env)	\
  __VClosureHead()				\
    __VClosureArgs3(arg1,arg2,arg3)		\
       __VClosureTail(name,body,env)

#define VClosure4(name,arg1,arg2,arg3,arg4,body,env)	\
  __VClosureHead()					\
    __VClosureArgs4(arg1,arg2,arg3,arg4)			\
       __VClosureTail(name,body,env)

#define VClosure5(name,arg1,arg2,arg3,arg4,arg5,body,env)	\
  __VClosureHead()						\
    __VClosureArgs5(arg1,arg2,arg3,arg4,arg5)			\
       __VClosureTail(name,body,env)

#define VClArg(name)             m_pparm[name]
#define VClReturn(val)           m_result = val;
#define VClResult(closure_name)  closure_name.m_result

#endif // __TCLOSURE_H__

// -- end of TClosure.h -- //
